"""Functional utilities for compilations."""

from collections.abc import Mapping
from functools import partial
from typing import Any, Callable, Iterable, Iterator, TypeVar

from preacher.compilation.error import CompilationError, on_key, on_index

T = TypeVar("T")
U = TypeVar("U")


def map_compile(func: Callable[[T], U], items: Iterable[T]) -> Iterator[U]:
    for idx, item in enumerate(items):
        with on_index(idx):
            yield func(item)


def run_recursively(func: Callable[[object], Any], obj) -> object:
    if isinstance(obj, Mapping):

        def _func(key: object, value: object) -> object:
            if not isinstance(key, str):
                message = f"Key must be a string, given {type(key)}: {key}"
                raise CompilationError(message)
            with on_key(key):
                return run_recursively(func, value)

        return {k: _func(k, v) for (k, v) in obj.items()}

    if isinstance(obj, list):
        _func = partial(run_recursively, func)
        return list(map_compile(_func, obj))
    return func(obj)


def compile_flattening(
    func: Callable[[object], T],
    obj: object,
) -> Iterator[T]:
    """Compile while flattening object, which can be a nested list."""

    if not isinstance(obj, list):
        yield func(obj)
        return

    for idx, item in enumerate(obj):
        with on_index(idx):
            yield from compile_flattening(func, item)



from .case import Case
from .case_listener import CaseListener
from .case_result import CaseResult
from .case_runner import CaseRunner
from .scenario import Scenario
from .scenario_result import ScenarioResult
from .scenario_runner import ScenarioRunner
from .scenario_task import ScenarioTask

__all__ = [
    "Case",
    "CaseRunner",
    "CaseListener",
    "CaseResult",
    "Scenario",
    "ScenarioRunner",
    "ScenarioResult",
    "ScenarioTask",
]



"""Date and time compilation."""

from datetime import timedelta

from preacher.core.datetime import DatetimeFormat, parse_timedelta
from preacher.core.datetime import ISO8601
from preacher.core.datetime import StrftimeFormat
from .error import CompilationError
from .util.type import ensure_optional_str, ensure_str


def compile_datetime_format(obj: object) -> DatetimeFormat:
    """
    Args:
        obj: The compiled value, which should be `None` or a string.
    Raises:
        CompilationError: When compilation fails.
    """
    format_string = ensure_optional_str(obj)
    if format_string is None or format_string.lower() == "iso8601":
        return ISO8601
    return StrftimeFormat(format_string)


def compile_timedelta(obj: object) -> timedelta:
    """
    Args:
        obj: The compiled value, which should be a string.
    Raises:
        CompilationError: When compilation fails.
    """
    value = ensure_str(obj)
    try:
        return parse_timedelta(value)
    except ValueError as error:
        raise CompilationError(str(error), cause=error)



from concurrent.futures import Executor
from typing import Optional

from preacher.core.request import Requester
from preacher.core.scenario import CaseRunner, ScenarioRunner
from preacher.core.unit import UnitRunner
from .listener import Listener
from .scenario_scheduler import ScenarioScheduler


def create_scheduler(
    executor: Executor,
    base_url: str = "",
    timeout: Optional[float] = None,
    retry: int = 0,
    delay: float = 0.1,
    listener: Optional[Listener] = None,
) -> ScenarioScheduler:
    requester = Requester(base_url=base_url, timeout=timeout)
    unit_runner = UnitRunner(requester=requester, retry=retry, delay=delay)
    case_runner = CaseRunner(unit_runner=unit_runner, listener=listener)
    runner = ScenarioRunner(executor=executor, case_runner=case_runner)
    return ScenarioScheduler(runner=runner, listener=listener)



"""Extraction package."""

from preacher.core.extraction.analysis import Analyzer
from preacher.core.extraction.analysis import MappingAnalyzer
from preacher.core.extraction.analysis import ResponseBodyAnalyzer
from preacher.core.extraction.error import ExtractionError
from preacher.core.extraction.extraction import Extractor

__all__ = [
    "Analyzer",
    "ResponseBodyAnalyzer",
    "MappingAnalyzer",
    "ExtractionError",
    "Extractor",
]



from collections import deque
from datetime import time, datetime, timedelta
from typing import Optional, Type

from dateutil.parser import isoparse

from preacher.core.context import Context, CONTEXT_KEY_STARTS
from preacher.core.datetime import DatetimeFormat
from preacher.core.datetime import DatetimeWithFormat
from preacher.core.datetime import ISO8601
from preacher.core.datetime import now
from preacher.core.datetime import parse_time
from preacher.core.datetime import parse_timedelta
from preacher.core.datetime import system_timezone
from preacher.core.value.value import Value
from .static import StaticValue


class RelativeDatetime(Value[datetime]):
    def __init__(self, delta: Optional[timedelta] = None, tm: Optional[time] = None):
        self._delta = delta or timedelta()
        self._tm = tm

    @property
    def type(self) -> Type[datetime]:
        return datetime

    def resolve(self, context: Optional[Context] = None) -> datetime:
        origin = _select_origin(context)
        resolved = origin + self._delta
        if self._tm:
            resolved = datetime.combine(resolved.date(), self._tm)
        return resolved


class DatetimeValueWithFormat(Value[DatetimeWithFormat]):
    def __init__(self, original: Value[datetime], fmt: Optional[DatetimeFormat] = None):
        self._original = original
        self._fmt = fmt or ISO8601

    @property
    def type(self) -> Type[DatetimeWithFormat]:
        return DatetimeWithFormat

    def resolve(self, context: Optional[Context] = None) -> DatetimeWithFormat:
        resolved = self._original.resolve(context)
        return DatetimeWithFormat(resolved, self._fmt)


def parse_datetime_value_with_format(
    value: object,
    fmt: Optional[DatetimeFormat] = None,
) -> Value[DatetimeWithFormat]:
    """
    Args:
        value: The compiled value, which should be a datetime or a string.
        fmt: The datetime format.
    Raises:
        ValueError: When parsing fails.
    """
    if isinstance(value, datetime):
        if not value.tzinfo:
            value = value.replace(tzinfo=system_timezone())
        return StaticValue(DatetimeWithFormat(value, fmt))

    # Try to parse `obj` as a datetime-compatible string below.
    if not isinstance(value, str):
        raise ValueError(f"Must be a datetime-compatible value, but given {type(value)}: {value}")
    relative_datetime = parse_relative_datetime_value(value)
    return DatetimeValueWithFormat(relative_datetime, fmt)


def parse_relative_datetime_value(value: str) -> RelativeDatetime:
    delta: timedelta = timedelta()
    tm: Optional[time] = None

    words = deque(value.split())
    while words:
        word = words.popleft()

        # time
        try:
            tm = parse_time(word)
            continue
        except ValueError:
            pass  # Try to compile value as another format.

        # 1 word timedelta
        try:
            delta += parse_timedelta(word)
            continue
        except ValueError:
            pass  # Try to compile value as another format.

        # 2 words timedelta
        if not words:
            raise ValueError(f"Invalid format: {value}")
        word += words.popleft()
        try:
            delta += parse_timedelta(word)
        except ValueError:
            raise ValueError(f"Invalid format: {value}")

    return RelativeDatetime(delta, tm)


def _select_origin(context: Optional[Context] = None) -> datetime:
    if context is None:
        return now()

    origin = context.get(CONTEXT_KEY_STARTS)
    if not origin:
        return now()
    if isinstance(origin, datetime):
        return origin
    if isinstance(origin, str):
        return isoparse(origin)
    raise ValueError(f"Invalid datetime value: {origin}")



import contextlib
import io
import logging
import sys
import uuid
from typing import Iterable, Iterator, Optional

from preacher.core.request import ExecutionReport
from preacher.core.scenario import ScenarioResult, CaseResult
from preacher.core.status import Status
from preacher.core.verification import ResponseVerification, Verification

_LEVEL_MAP = {
    Status.SKIPPED: logging.DEBUG,
    Status.SUCCESS: logging.INFO,
    Status.UNSTABLE: logging.WARN,
    Status.FAILURE: logging.ERROR,
}


class LoggingReporter:
    def __init__(self, logger: logging.Logger):
        self._logger = logger
        self._indent = ""

    def show_scenario_result(self, scenario: ScenarioResult) -> None:
        status = scenario.status
        level = _LEVEL_MAP[status]

        label = scenario.label or "Not labeled scenario"
        self._log(level, "%s: %s", label, status)

        message = scenario.message
        if message:
            with self._nesting():
                self._multi_line_message(level, message)

        with self._nesting():
            for case in scenario.cases.items:
                self.show_case_result(case)

            for subscenario in scenario.subscenarios.items:
                self.show_scenario_result(subscenario)

    def show_case_result(self, case: CaseResult) -> None:
        status = case.status
        level = _LEVEL_MAP[status]

        label = case.label or "Not labeled case"
        self._log(level, "%s: %s", label, status)
        with self._nesting():
            self.show_execution(case.execution)

            response = case.response
            if response:
                self.show_response_verification(response)

    def show_execution(self, execution: ExecutionReport) -> None:
        status = execution.status
        level = _LEVEL_MAP[status]

        self._log(level, "Execution: %s", status)
        if execution.message:
            with self._nesting():
                self._multi_line_message(level, execution.message)

    def show_response_verification(
        self,
        verification: ResponseVerification,
        label: str = "Response",
    ) -> None:
        status = verification.status
        level = _LEVEL_MAP[status]
        self._log(level, "%s: %s", label, status)
        with self._nesting():
            self.show_verification(
                verification=verification.status_code,
                label="Status Code",
            )
            self.show_verification(
                verification=verification.headers,
                label="Headers",
                child_label="Description",
            )
            self.show_verification(
                verification=verification.body,
                label="Body",
                child_label="Description",
            )

    def show_verification(
        self,
        verification: Verification,
        label: str,
        child_label: str = "Predicate",
    ) -> None:
        status = verification.status
        level = _LEVEL_MAP[status]
        self._log(level, "%s: %s", label, status)
        message = verification.message
        if message:
            with self._nesting():
                self._multi_line_message(level, message)

        with self._nesting():
            for idx, child in enumerate(verification.children):
                self.show_verification(child, f"{child_label} {idx + 1}")

    def show_status(self, status: Status) -> None:
        level = _LEVEL_MAP[status]
        self._log(level, "%s", status)

    def _log(self, level: int, message: str, *args) -> None:
        self._logger.log(level, self._indent + message, *args)

    def _multi_line_message(self, level: int, message: str) -> None:
        for line in io.StringIO(message):
            self._log(level, line.rstrip())

    @contextlib.contextmanager
    def _nesting(self) -> Iterator[None]:
        original = self._indent
        self._indent += ".."
        yield
        self._indent = original


def create_logging_reporter(
    logger: Optional[logging.Logger] = None,
    logger_name: str = "",
    level: Status = Status.SUCCESS,
    handlers: Optional[Iterable[logging.Handler]] = None,
    formatter: Optional[logging.Formatter] = None,
) -> LoggingReporter:
    """
    Create a logging reporter.

    Args:
        logger: A logger where reports logged. When given this, the other parameters are ignored.
        logger_name: The logger name. When not given, it will be automatically generated.
        level: The minimum level to report.
        handlers: The logging handlers. When given, `formatter` is ignored.
        formatter: The logging formatter.
    """
    if not logger:
        logging_level = _status_to_logging_level(level)
        logger = logging.getLogger(logger_name or str(uuid.uuid4()))
        logger.setLevel(logging_level)

        if not handlers:
            default_handler = logging.StreamHandler(sys.stdout)
            default_handler.setLevel(logging_level)
            if formatter:
                default_handler.setFormatter(formatter)
            handlers = (default_handler,)
        for handler in handlers:
            logger.addHandler(handler)

    return LoggingReporter(logger)


def _status_to_logging_level(level: Status) -> int:
    if level is Status.SKIPPED:
        return logging.DEBUG
    if level is Status.SUCCESS:
        return logging.INFO
    if level is Status.UNSTABLE:
        return logging.WARNING
    return logging.ERROR



"""
Descriptions, which extract a value from an analyzer and test
along the given predicates.
"""

from typing import List, Optional

from preacher.core.context import Context
from preacher.core.extraction import Analyzer, Extractor
from .predicate import Predicate
from .verification import Verification


class Description:
    def __init__(
        self,
        extractor: Extractor,
        predicates: List[Predicate],
        value_name: Optional[str] = None,
    ):
        self._extractor = extractor
        self._predicates = predicates
        self._value_name = value_name

    def verify(self, analyzer: Analyzer, context: Optional[Context] = None) -> Verification:
        try:
            value = self._extractor.extract(analyzer)
        except Exception as error:
            return Verification.of_error(error)

        if self._value_name and context is not None:
            context[self._value_name] = value

        return Verification.collect(
            predicate.verify(value, context) for predicate in self._predicates
        )



from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar, Type

from preacher.core.context import Context

T = TypeVar("T")


class Value(ABC, Generic[T]):
    @property
    @abstractmethod
    def type(self) -> Type[T]:
        ...  # pragma: no cover

    @abstractmethod
    def resolve(self, context: Optional[Context] = None) -> T:
        ...  # pragma: no cover



"""Matcher compilation."""

from collections.abc import Mapping
from typing import Callable, Dict, Iterable, Tuple, Union

import hamcrest
from hamcrest.core.matcher import Matcher

from preacher.compilation.error import CompilationError, on_key
from preacher.compilation.util.functional import map_compile
from preacher.compilation.util.type import ensure_list
from preacher.core.value import Value
from preacher.core.value.impl.datetime import parse_datetime_value_with_format
from preacher.core.value.impl.static import StaticValue
from preacher.core.verification.hamcrest import before, after, day_of_week
from preacher.core.verification.matcher import MatcherFactory
from preacher.core.verification.matcher import MatcherFunc
from preacher.core.verification.matcher import RecursiveMatcherFactory
from preacher.core.verification.matcher import StaticMatcherFactory
from preacher.core.verification.matcher import ValueMatcherFactory
from preacher.core.verification.type import require_type

ValueFunc = Callable[[object], Value]


class MatcherFactoryCompiler:
    """
    A matcher factory compiler.
    """

    _DEFAULT_VALUE_FUNC = StaticValue

    def __init__(self):
        self._static: Dict[str, MatcherFactory] = {}
        self._taking_value: Dict[str, Tuple[MatcherFunc, ValueFunc]] = {}
        self._recursive: Dict[str, Tuple[MatcherFunc, bool]] = {}

    def add_static(
        self,
        keys: Union[str, Iterable[str]],
        item: Union[MatcherFactory, Matcher],
    ) -> None:
        """
        Add a static matcher on given keys.

        Args:
            keys: The key(s) of the matcher.
            item: A matcher or a ``MatcherFactory`` to assign.
        """

        if isinstance(item, Matcher):
            item = StaticMatcherFactory(item)
        for key in self._ensure_keys(keys):
            self._static[key] = item

    def add_taking_value(
        self,
        keys: Union[str, Iterable[str]],
        matcher_func: MatcherFunc,
        value_func: ValueFunc = _DEFAULT_VALUE_FUNC,
    ) -> None:
        """
        Add a matcher taking a value on given keys.

        Args:
            keys: The key(s) of the matcher.
            matcher_func: A function that takes an object and returns a matcher.
            value_func (optional): A function that takes an object and returns a ``Value`` object.
        """

        for key in self._ensure_keys(keys):
            self._taking_value[key] = (matcher_func, value_func)

    def add_recursive(
        self,
        keys: Union[str, Iterable[str]],
        matcher_func: MatcherFunc,
        multiple: bool = True,
    ) -> None:
        """
        Add a matcher taking one or more matchers.

        Args:
            keys: The key(s) of the matcher.
            matcher_func: A function that takes one or more matchers and returns a matcher.
            multiple: Whether the matcher can take multiple arguments or not.
        """

        for key in self._ensure_keys(keys):
            self._recursive[key] = (matcher_func, multiple)

    def compile(self, obj: object) -> MatcherFactory:
        """
        Compile an object into a matcher factory.

        Args:
            obj: A compiled object.
        Returns:
            The result of compilation.
        Raises:
            CompilationError: when compilation fails.
        """

        if isinstance(obj, str) and obj in self._static:
            return self._static[obj]

        if isinstance(obj, Mapping):
            if len(obj) != 1:
                message = f"Must have only 1 element, but has {len(obj)}"
                raise CompilationError(message)

            key, value_obj = next(iter(obj.items()))
            if key in self._taking_value:
                return self._compile_taking_value(key, value_obj)
            if key in self._recursive:
                return self._compile_recursive(key, value_obj)

        return ValueMatcherFactory(hamcrest.equal_to, StaticValue(obj))

    def _compile_taking_value(self, key: str, obj: object):
        matcher_func, value_func = self._taking_value[key]
        return ValueMatcherFactory(matcher_func, arg=obj, value_func=value_func)

    def _compile_recursive(self, key: str, obj: object):
        matcher_func, multiple = self._recursive[key]
        if multiple:
            objs = ensure_list(obj)
            inner_matchers = list(map_compile(self.compile, objs))
        else:
            with on_key(key):
                inner_matchers = [self.compile(obj)]
        return RecursiveMatcherFactory(matcher_func, inner_matchers)

    @staticmethod
    def _ensure_keys(keys: Union[str, Iterable[str]]) -> Iterable[str]:
        if isinstance(keys, str):
            return (keys,)
        return keys


def add_default_matchers(compiler: MatcherFactoryCompiler) -> None:
    """
    Add default matchers to a compiler.

    Args:
        compiler: A compiler to be modified.
    """

    compiler.add_recursive(("be",), hamcrest.is_, multiple=False)

    # For objects.
    compiler.add_static(("be_null",), hamcrest.none())
    compiler.add_static(("not_be_null",), hamcrest.not_none())
    compiler.add_static(("be_monday",), day_of_week(0))
    compiler.add_static(("be_tuesday",), day_of_week(1))
    compiler.add_static(("be_wednesday",), day_of_week(2))
    compiler.add_static(("be_thursday",), day_of_week(3))
    compiler.add_static(("be_friday",), day_of_week(4))
    compiler.add_static(("be_saturday",), day_of_week(5))
    compiler.add_static(("be_sunday",), day_of_week(6))
    compiler.add_taking_value(("equal",), hamcrest.equal_to)
    compiler.add_recursive(("have_length",), hamcrest.has_length, multiple=False)

    # For comparable values.
    compiler.add_taking_value(("be_greater_than",), hamcrest.greater_than)
    compiler.add_taking_value(("be_greater_than_or_equal_to",), hamcrest.greater_than_or_equal_to)
    compiler.add_taking_value(("be_less_than",), hamcrest.less_than)
    compiler.add_taking_value(("be_less_than_or_equal_to",), hamcrest.less_than_or_equal_to)

    # For strings.
    compiler.add_taking_value(("contain_string",), require_type(str, hamcrest.contains_string))
    compiler.add_taking_value(("start_with",), require_type(str, hamcrest.starts_with))
    compiler.add_taking_value(("end_with",), require_type(str, hamcrest.ends_with))
    compiler.add_taking_value(("match_regexp",), require_type(str, hamcrest.matches_regexp))

    # For collections.
    compiler.add_recursive(("have_item",), hamcrest.has_item, multiple=False)
    compiler.add_recursive(("have_items",), hamcrest.has_items)
    compiler.add_recursive(("contain_exactly",), hamcrest.contains_exactly)
    compiler.add_recursive(("contain_in_any_order",), hamcrest.contains_inanyorder)

    # For datetime.
    compiler.add_taking_value(("be_before",), before, parse_datetime_value_with_format)
    compiler.add_taking_value(("be_after",), after, parse_datetime_value_with_format)

    # For collections.
    compiler.add_static(("be_empty",), StaticMatcherFactory(hamcrest.empty()))

    # Logical.
    compiler.add_static("anything", StaticMatcherFactory(hamcrest.anything()))
    compiler.add_recursive(("not",), hamcrest.not_, multiple=False)
    compiler.add_recursive(("all_of",), hamcrest.all_of)
    compiler.add_recursive(("any_of",), hamcrest.any_of)



from yaml import Node
from yamlen import Tag, TagContext

from preacher.compilation.argument import Argument


class ArgumentTag(Tag):
    def construct(self, node: Node, context: TagContext) -> object:
        key = context.constructor.construct_scalar(node)  # type: ignore
        return Argument(key)



"""Preacher plugin definitions."""

from pluggy import PluginManager

from . import hookspec

__all__ = ["get_plugin_manager"]


def get_plugin_manager() -> PluginManager:
    """
    Get a plugin manager that the entry points and the default implementation are already loaded.
    """
    manager = PluginManager("preacher")
    manager.add_hookspecs(hookspec)
    manager.load_setuptools_entrypoints("preacher")

    return manager



from typing import Optional

from preacher.core.request import ExecutionReport, Response


class CaseListener:
    """
    Interface to listen to running cases.
    Default implementations do nothing.
    """

    def on_execution(self, execution: ExecutionReport, response: Optional[Response]) -> None:
        pass



from typing import Iterator

from preacher.core.extraction import ExtractionError
from preacher.core.extraction.impl.jq_ import JqEngine


class PyJqEngine(JqEngine):
    def __init__(self):
        import jq

        self._compile = jq.compile

    def iter(self, query: str, text: str) -> Iterator[object]:
        try:
            compiled = self._compile(query)
        except ValueError:
            raise ExtractionError(f"Invalid jq script: {query}")
        return compiled.input(text=text)

    @staticmethod
    def is_available() -> bool:
        try:
            import jq  # noqa: F401

            return True
        except ImportError:  # pragma: no cover
            return False



"""Verification."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Optional, Sequence

from preacher.core.status import Status, Statused, merge_statuses
from preacher.core.util.error import to_message


@dataclass(frozen=True)
class Verification(Statused):
    status: Status = Status.SKIPPED
    message: Optional[str] = None
    children: Sequence[Verification] = ()

    @staticmethod
    def succeed() -> Verification:
        return Verification(status=Status.SUCCESS)

    @staticmethod
    def of_error(error: Exception) -> Verification:
        return Verification(status=Status.FAILURE, message=to_message(error))

    @staticmethod
    def collect(children: Iterable[Verification]) -> Verification:
        children = list(children)
        status = merge_statuses(child.status for child in children)
        return Verification(status=status, children=children)



"""Matchers."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable, List, Optional, TypeVar

from hamcrest import assert_that
from hamcrest.core.matcher import Matcher

from preacher.core.context import Context
from preacher.core.status import Status
from preacher.core.value import Value
from preacher.core.value.impl.static import StaticValue
from .predicate import Predicate
from .verification import Verification

T = TypeVar("T")
MatcherFunc = Callable[..., Matcher]


class MatcherWrappingPredicate(Predicate):
    """Matcher implemented by hamcrest matchers."""

    def __init__(self, factory: MatcherFactory):
        self._factory = factory

    def verify(self, actual: object, context: Optional[Context] = None) -> Verification:
        try:
            hamcrest_matcher = self._factory.create(context)
            assert_that(actual, hamcrest_matcher)
        except AssertionError as error:
            message = str(error).strip()
            return Verification(status=Status.UNSTABLE, message=message)
        except Exception as error:
            return Verification.of_error(error)

        return Verification.succeed()


class MatcherFactory(ABC):
    @abstractmethod
    def create(self, context: Optional[Context] = None) -> Matcher:
        ...  # pragma: no cover


class StaticMatcherFactory(MatcherFactory):
    def __init__(self, matcher: Matcher):
        self._matcher = matcher

    def create(self, context: Optional[Context] = None) -> Matcher:
        return self._matcher


class ValueMatcherFactory(MatcherFactory):
    def __init__(
        self,
        matcher_func: MatcherFunc,
        arg: object,
        value_func: Callable[[object], Value] = StaticValue,
    ):
        self._inner_factory = matcher_func
        self._arg = arg
        self._value_func = value_func

    def create(self, context: Optional[Context] = None) -> Matcher:
        resolved_value = self._ensure_value().resolve(context)
        return self._inner_factory(resolved_value)

    def _ensure_value(self) -> Value:
        if isinstance(self._arg, Value):
            return self._arg
        return self._value_func(self._arg)


class RecursiveMatcherFactory(MatcherFactory):
    def __init__(self, matcher_func: MatcherFunc, inner_factories: List[MatcherFactory]):
        self._matcher_func = matcher_func
        self._inner_factories = inner_factories

    def create(self, context: Optional[Context] = None) -> Matcher:
        inner_matchers = (factory.create(context) for factory in self._inner_factories)
        return self._matcher_func(*inner_matchers)



"""Typing utilities."""

from typing import Mapping, Optional, TypeVar

from preacher.compilation.error import CompilationError

T = TypeVar("T")


def ensure_bool(obj: object) -> bool:
    """
    Ensure a boolean object.

    Args:
        obj: An ensured object, which should be a `bool` value.
    Returns:
        The compiled value.
    Raises:
        CompilationError: when compilation fails.
    """
    if not isinstance(obj, bool):
        raise CompilationError(f"Must be a boolean, given {type(obj)}")
    return obj


def ensure_str(obj: object) -> str:
    """
    Ensure a string object.

    Args:
        obj: An ensured object, which should be a `string` value.
    Returns:
        The compiled value.
    Raises:
        CompilationError: when compilation fails.
    """
    if not isinstance(obj, str):
        raise CompilationError(f"must be a string, given {type(obj)}")
    return obj


def ensure_optional_str(obj: object) -> Optional[str]:
    """
    Compile an optional string object.

    Args:
        obj: An ensured object, which should be a `string` value or `None`.
    Returns:
        The compiled value.
    Raises:
        CompilationError: when compilation fails.
    """
    if obj is None:
        return None
    return ensure_str(obj)


def ensure_list(obj: object) -> list:
    """
    Ensure a list object.
    When given not a list, then this returns ``[it]``.

    Args:
        obj: A ensured object.
    Returns:
        The compilation result.
    """
    if not isinstance(obj, list):
        return [obj]
    return obj


def ensure_mapping(obj: object) -> Mapping:
    """
    Ensure a mapping object.

    Args:
        obj: A compiled object, which should be a mapping.
    Returns:
        The compile result.
    Raises:
        CompilationError: when compilation fails.
    """
    if not isinstance(obj, Mapping):
        raise CompilationError(f"Must be a map, given {type(obj)}")
    return obj


def or_else(optional: Optional[T], default: T) -> T:
    if optional is None:
        return default
    return optional



"""Extraction compilation."""

from typing import Any, Callable, Mapping, Optional

from preacher.compilation.error import CompilationError, on_key
from preacher.compilation.util.type import ensure_bool, ensure_mapping, ensure_str
from preacher.core.extraction.extraction import Extractor
from preacher.core.extraction.impl.key import KeyExtractor
from preacher.core.extraction.impl.jq_ import JqExtractor
from preacher.core.extraction.impl.jq_engine import PyJqEngine
from preacher.core.extraction.impl.xpath import XPathExtractor

_CAST_FUNC_MAP = {
    "int": int,
    "float": float,
    "string": str,
}
_KEY_JQ = "jq"
_KEY_XPATH = "xpath"
_KEY_KEY = "key"
_KEY_MULTIPLE = "multiple"
_KEY_CAST_TO = "cast_to"
_KEY_NAMESPACES = "namespaces"


Factory = Callable[[str, Mapping[str, object]], Extractor]


class ExtractionCompiler:
    def __init__(self):
        self._factory_map = {}

    def add_factory(self, key: str, factory: Factory):
        self._factory_map[key] = factory

    def compile(self, obj: object) -> Extractor:
        """`obj` should be a mapping or a string."""

        if isinstance(obj, str):
            return self.compile({_KEY_JQ: obj})

        if not isinstance(obj, Mapping):
            message = f"Must be a map or a string, given {type(obj)}"
            raise CompilationError(message)

        key_candidates = self._factory_map.keys() & obj.keys()
        key_count = len(key_candidates)
        if key_count != 1:
            raise CompilationError(f"Must have only 1 extraction key, but has {key_count}")
        factory_key = next(iter(key_candidates))
        factory = self._factory_map[factory_key]
        query = obj[factory_key]

        return factory(query, obj)


def _select_multiple(options: Mapping) -> bool:
    obj = options.get(_KEY_MULTIPLE, False)
    with on_key(_KEY_MULTIPLE):
        return ensure_bool(obj)


def _select_cast(options: Mapping) -> Optional[Callable[[object], Any]]:
    obj = options.get(_KEY_CAST_TO)
    if obj is None:
        return None

    with on_key(_KEY_CAST_TO):
        key = ensure_str(obj)
        cast = _CAST_FUNC_MAP.get(key)
        if not cast:
            raise CompilationError(f"Invalid value: {key}")

        return cast


def compile_jq(query: str, options: Mapping) -> JqExtractor:
    multiple = _select_multiple(options)
    cast = _select_cast(options)
    return JqExtractor(PyJqEngine(), query, multiple=multiple, cast=cast)


def _ensure_str_on_key(key: str, obj: object) -> str:
    with on_key(key):
        return ensure_str(obj)


def _select_namespaces(options: Mapping) -> Mapping[str, str]:
    obj = options.get(_KEY_NAMESPACES, {})
    with on_key(_KEY_NAMESPACES):
        mapping = ensure_mapping(obj)
        return {ensure_str(key): _ensure_str_on_key(key, value) for key, value in mapping.items()}


def compile_xpath(query: str, options: Mapping) -> XPathExtractor:
    multiple = _select_multiple(options)
    cast = _select_cast(options)
    namespaces = _select_namespaces(options)

    return XPathExtractor(query, multiple=multiple, cast=cast, namespaces=namespaces)


def compile_key(query: str, options: Mapping) -> KeyExtractor:
    multiple = _select_multiple(options)
    cast = _select_cast(options)
    return KeyExtractor(query, multiple=multiple, cast=cast)


def add_default_extractions(compiler: ExtractionCompiler) -> None:
    if PyJqEngine.is_available():
        compiler.add_factory(_KEY_JQ, compile_jq)
    compiler.add_factory(_KEY_XPATH, compile_xpath)
    compiler.add_factory(_KEY_KEY, compile_key)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, replace
from typing import Mapping, Optional, Callable

from preacher.compilation.argument import Arguments, inject_arguments
from preacher.compilation.error import CompilationError, on_key
from preacher.compilation.util.type import ensure_str, ensure_mapping
from preacher.core.request import (
    RequestBody,
    UrlencodedRequestBody,
    UrlParams,
    JsonRequestBody,
)
from .url_param import compile_url_params

_NOT_SPECIFIED = object()
_KEY_TYPE = "type"
_KEY_DATA = "data"


class RequestBodyCompiled(ABC):
    @abstractmethod
    def replace(self, other: RequestBodyCompiled) -> RequestBodyCompiled:
        ...  # pragma: no cover

    @abstractmethod
    def compile_and_replace(self, obj: Mapping) -> RequestBodyCompiled:
        ...  # pragma: no cover

    @abstractmethod
    def fix(self) -> RequestBody:
        ...  # pragma: no cover


@dataclass(frozen=True)
class UrlencodedRequestBodyCompiled(RequestBodyCompiled):
    data: Optional[UrlParams] = None

    def replace(self, other: RequestBodyCompiled) -> RequestBodyCompiled:
        if not isinstance(other, UrlencodedRequestBodyCompiled):
            return other
        return replace(
            self,
            data=other.data if other.data is not None else self.data,
        )

    def compile_and_replace(self, obj: Mapping) -> RequestBodyCompiled:
        compiled = self

        data_obj = obj.get(_KEY_DATA)
        if data_obj is not None:
            with on_key(_KEY_DATA):
                data = compile_url_params(data_obj)
            compiled = replace(self, data=data)

        return compiled

    def fix(self) -> RequestBody:
        return UrlencodedRequestBody(params=self.data or {})


@dataclass(frozen=True)
class JsonRequestBodyCompiled(RequestBodyCompiled):
    data: object = _NOT_SPECIFIED

    def replace(self, other: RequestBodyCompiled) -> RequestBodyCompiled:
        if not isinstance(other, JsonRequestBodyCompiled):
            return other
        return replace(
            self,
            data=other.data if other.data is not _NOT_SPECIFIED else self.data,
        )

    def compile_and_replace(self, obj: Mapping) -> RequestBodyCompiled:
        compiled = self

        if _KEY_DATA in obj:
            data = obj[_KEY_DATA]
            compiled = replace(self, data=data)

        return compiled

    def fix(self) -> RequestBody:
        return JsonRequestBody(
            data=self.data if self.data is not _NOT_SPECIFIED else None,
        )


_TYPE_FACTORY_MAP: Mapping[str, Callable[[], RequestBodyCompiled]] = {
    "urlencoded": UrlencodedRequestBodyCompiled,
    "json": JsonRequestBodyCompiled,
}


class RequestBodyCompiler:
    def __init__(self, default: Optional[RequestBodyCompiled] = None):
        self._default: RequestBodyCompiled = default or UrlencodedRequestBodyCompiled()

    def compile(
        self,
        obj: object,
        arguments: Optional[Arguments] = None,
    ) -> RequestBodyCompiled:
        """
        Compiles an object into an intermediate request body.

        Args:
            obj: A compiled object, which should be a mapping.
            arguments: Arguments to inject.
        Returns:
            The result of compilation.
        Raises:
            CompilationError: when compilation fails.
        """

        obj = inject_arguments(obj, arguments)
        obj = ensure_mapping(obj)
        compiled = self._default

        type_obj = obj.get(_KEY_TYPE)
        if type_obj is not None:
            with on_key(_KEY_TYPE):
                key = ensure_str(type_obj)
                factory = _TYPE_FACTORY_MAP.get(key)
                if not factory:
                    raise CompilationError(
                        f"Must be in {list(_TYPE_FACTORY_MAP)}" f", but given: {key}"
                    )
            compiled = self._default.replace(factory())

        return compiled.compile_and_replace(obj)

    def of_default(self, default: RequestBodyCompiled) -> RequestBodyCompiler:
        """
        Creates a new compiler
        that is configured of the default request body.

        Args:
            default: A default request body.
        Returns:
            A new compiler.
        """

        return RequestBodyCompiler(default=self._default.replace(default))



from concurrent.futures import Executor

from preacher.core.context import Context, CONTEXT_KEY_BASE_URL, CONTEXT_KEY_STARTS
from preacher.core.datetime import now
from preacher.core.extraction import MappingAnalyzer
from preacher.core.status import Status
from preacher.core.verification import Verification
from .case_runner import CaseRunner
from .scenario import Scenario
from .scenario_result import ScenarioResult
from .scenario_task import ScenarioTask, StaticScenarioTask, RunningScenarioTask
from .util.concurrency import CasesTask, OrderedCasesTask, UnorderedCasesTask


class ScenarioRunner:
    def __init__(self, executor: Executor, case_runner: CaseRunner):
        self._executor = executor
        self._case_runner = case_runner

    def submit(self, scenario: Scenario) -> ScenarioTask:
        starts = now()
        context = Context(
            **{
                CONTEXT_KEY_STARTS: starts,
                CONTEXT_KEY_BASE_URL: self._case_runner.base_url,
            }
        )

        context_analyzer = MappingAnalyzer(context)
        conditions = Verification.collect(
            condition.verify(context_analyzer, context) for condition in scenario.conditions
        )
        if not conditions.status.is_succeeded:
            status = Status.SKIPPED
            if conditions.status is Status.FAILURE:
                status = Status.FAILURE

            result = ScenarioResult(label=scenario.label, status=status, conditions=conditions)
            return StaticScenarioTask(result)

        if scenario.ordered:
            cases: CasesTask = OrderedCasesTask(
                self._executor,
                self._case_runner,
                scenario.cases,
                context=context,
            )
        else:
            cases = UnorderedCasesTask(self._executor, self._case_runner, scenario.cases)
        subscenarios = [self.submit(subscenario) for subscenario in scenario.subscenarios]
        return RunningScenarioTask(
            label=scenario.label,
            conditions=conditions,
            cases=cases,
            subscenarios=subscenarios,
        )



from .description import DescriptionCompiler
from .factory import create_description_compiler
from .factory import create_predicate_compiler
from .factory import create_response_description_compiler
from .matcher import MatcherFactoryCompiler
from .predicate import PredicateCompiler
from .response import ResponseDescriptionCompiler, ResponseDescriptionCompiled

__all__ = [
    "DescriptionCompiler",
    "MatcherFactoryCompiler",
    "PredicateCompiler",
    "ResponseDescriptionCompiler",
    "ResponseDescriptionCompiled",
    "create_predicate_compiler",
    "create_description_compiler",
    "create_response_description_compiler",
]



import uuid
from copy import copy
from dataclasses import dataclass, field, replace
from datetime import datetime
from typing import Mapping, Union, Optional, Tuple

import requests

from preacher import __version__ as _version
from preacher.core.context import Context, closed_context
from preacher.core.datetime import now
from preacher.core.status import Statused, Status
from preacher.core.util.error import to_message
from .request import Request
from .response import Response, ResponseBody
from .url_param import resolve_url_params

_DEFAULT_HEADERS = {"User-Agent": f"Preacher {_version}"}


class ResponseBodyWrapper(ResponseBody):
    def __init__(self, res: requests.Response):
        self._res = res

    @property
    def text(self) -> str:
        return self._res.text

    @property
    def content(self) -> bytes:
        return self._res.content


class ResponseWrapper(Response):
    def __init__(self, id: str, res: requests.Response):
        self._id = id
        self._res = res
        self._body = ResponseBodyWrapper(self._res)

    @property
    def id(self) -> str:
        return self._id

    @property
    def elapsed(self) -> float:
        return self._res.elapsed.total_seconds()

    @property
    def status_code(self) -> int:
        return self._res.status_code

    @property
    def headers(self) -> Mapping[str, str]:
        # Convert to the normal dictionary to adapt jq.
        # Names are converted to lower case to normalize.
        return {name.lower(): value for (name, value) in self._res.headers.items()}

    @property
    def body(self) -> ResponseBody:
        return self._body


@dataclass
class PreparedRequest:
    method: str
    url: str
    headers: Mapping[str, str]
    body: Union[None, str, bytes]


@dataclass(frozen=True)
class ExecutionReport(Statused):
    status: Status = Status.SKIPPED
    starts: datetime = field(default_factory=now)
    request: Optional[PreparedRequest] = None
    message: Optional[str] = None


class Requester:
    def __init__(
        self,
        base_url: str = "",
        timeout: Optional[float] = None,
    ):
        """
        Args:
            base_url: A base URL.
            timeout: The timeout in seconds. ``None`` means no timeout.
        """
        self._base_url = base_url
        self._timeout = timeout

    @property
    def base_url(self) -> str:
        return self._base_url

    def execute(
        self,
        request: Request,
        session: Optional[requests.Session] = None,
        context: Optional[Context] = None,
    ) -> Tuple[ExecutionReport, Optional[Response]]:
        """
        Executes a request.

        Args:
            request: A request.
            session: A session object to execute.
            context: Execution context.
        Returns:
            A tuple of execution report and response.
            When there is no response, the response will be ``None``.
        """
        if session is None:
            with requests.Session() as new_session:
                return self.execute(request, session=new_session)

        context = context if context is not None else Context()
        starts = now()
        report = ExecutionReport(starts=starts)
        try:
            with closed_context(context, starts=starts) as context:
                prepped = self._prepare_request(request, context)
            proxies = session.rebuild_proxies(prepped, proxies=None)
        except Exception as error:
            message = to_message(error)
            report = replace(report, status=Status.FAILURE, message=message)
            return report, None

        report = replace(
            report,
            request=PreparedRequest(
                method=prepped.method or "",
                url=prepped.url or "",
                headers=prepped.headers,
                body=prepped.body,
            ),
        )

        try:
            res = session.send(prepped, proxies=proxies, timeout=self._timeout)
        except Exception as error:
            message = to_message(error)
            report = replace(report, status=Status.UNSTABLE, message=message)
            return report, None

        report = replace(report, status=Status.SUCCESS)
        response = ResponseWrapper(id=_generate_id(), res=res)
        return report, response

    def _prepare_request(self, request: Request, context: Context) -> requests.PreparedRequest:
        url = self._base_url + request.path
        headers = copy(_DEFAULT_HEADERS)

        data = None
        if request.body:
            content_type = request.body.content_type
            headers["Content-Type"] = content_type
            data = request.body.resolve(context)

        headers.update(request.headers)
        params = resolve_url_params(request.params, context)

        req = requests.Request(
            method=request.method.value,
            url=url,
            headers=headers,
            params=params,
            data=data,
        )
        return req.prepare()


def _generate_id() -> str:
    return str(uuid.uuid4())



from typing import Optional, Type

from preacher.core.context import Context
from preacher.core.value import Value


class ContextualValue(Value[object]):
    def __init__(self, key: str):
        self._key = key

    @property
    def type(self) -> Type[object]:
        return object

    @property
    def key(self) -> str:
        return self._key

    def resolve(self, context: Optional[Context] = None) -> object:
        context = context or Context()
        return context.get(self._key)



from dataclasses import dataclass, field
from typing import Optional

from preacher.core.request import ExecutionReport
from preacher.core.status import Statused, Status, merge_statuses
from preacher.core.verification import Verification, ResponseVerification


@dataclass(frozen=True)
class CaseResult(Statused):
    """
    Results for the test cases.
    """

    label: Optional[str] = None
    conditions: Verification = field(default_factory=Verification)
    execution: ExecutionReport = field(default_factory=ExecutionReport)
    response: Optional[ResponseVerification] = None

    @property
    def status(self) -> Status:  # HACK: should be cached
        if self.conditions.status == Status.UNSTABLE:
            return Status.SKIPPED
        if self.conditions.status == Status.FAILURE:
            return Status.FAILURE

        return merge_statuses(
            [
                self.execution.status,
                self.response.status if self.response else Status.SKIPPED,
            ]
        )



from logging import Logger
from typing import Optional

from pluggy import PluginManager

from preacher.compilation.extraction import ExtractionCompiler
from preacher.core.logger import default_logger


def create_extraction_compiler(
    plugin_manager: Optional[PluginManager] = None,
    logger: Optional[Logger] = None,
) -> ExtractionCompiler:
    logger = logger or default_logger
    compiler = ExtractionCompiler()
    if plugin_manager:
        plugin_manager.hook.preacher_add_extractions(compiler=compiler)
    else:
        message = "No plugin manager is given. The default extraction methods are not available."
        logger.info(message)
    return compiler



"""
Functional utilities.
"""

from typing import Any, Callable, TypeVar

T = TypeVar("T")


def identity(value: T, *_args, **_kwargs) -> T:
    """Returns the first argument, ignoring extra arguments."""
    return value


def recursive_map(func: Callable, obj: object) -> object:
    """
    Map `func` recursively, which affects all items in lists
    and all values in dictionaries.
    """
    if isinstance(obj, dict):
        return {k: recursive_map(func, v) for (k, v) in obj.items()}
    if isinstance(obj, list):
        return [recursive_map(func, v) for v in obj]
    return func(obj)


def apply_if_not_none(func: Callable[[object], Any], value: object) -> Any:
    """Apply `func` if `value` is not `None`."""
    if value is None:
        return None
    return func(value)



"""Case compilation."""

from __future__ import annotations

from dataclasses import dataclass, replace
from datetime import timedelta
from typing import List, Optional

from preacher.compilation.datetime import compile_timedelta
from preacher.compilation.error import on_key
from preacher.compilation.request import RequestCompiler, RequestCompiled
from preacher.compilation.util.functional import map_compile
from preacher.compilation.util.type import (
    ensure_bool,
    ensure_optional_str,
    ensure_list,
    ensure_mapping,
    or_else,
)
from preacher.compilation.verification import (
    ResponseDescriptionCompiled,
    ResponseDescriptionCompiler,
    DescriptionCompiler,
)
from preacher.core.scenario import Case
from preacher.core.verification import Description

_KEY_LABEL = "label"
_KEY_ENABLED = "enabled"
_KEY_CONDITIONS = "when"
_KEY_REQUEST = "request"
_KEY_RESPONSE = "response"
_KEY_WAIT = "wait"


@dataclass(frozen=True)
class CaseCompiled:
    label: Optional[str] = None
    enabled: Optional[bool] = None
    conditions: Optional[List[Description]] = None
    request: Optional[RequestCompiled] = None
    response: Optional[ResponseDescriptionCompiled] = None
    wait: Optional[timedelta] = None

    def replace(self, other: CaseCompiled) -> CaseCompiled:
        return CaseCompiled(
            label=or_else(other.label, self.label),
            enabled=or_else(other.enabled, self.enabled),
            conditions=or_else(other.conditions, self.conditions),
            request=or_else(other.request, self.request),
            response=or_else(other.response, self.response),
            wait=or_else(other.wait, self.wait),
        )

    def fix(self) -> Case:
        return Case(
            label=self.label,
            enabled=or_else(self.enabled, True),
            conditions=self.conditions or [],
            request=self.request.fix() if self.request else None,
            response=self.response.fix() if self.response else None,
            waiting_time=self.wait,
        )


class CaseCompiler:
    def __init__(
        self,
        request: RequestCompiler,
        response: ResponseDescriptionCompiler,
        description: DescriptionCompiler,
        default: Optional[CaseCompiled] = None,
    ):
        self._request = request
        self._response = response
        self._description = description
        self._default = default or CaseCompiled()

    def compile_fixed(self, obj: object) -> Case:
        """
        Compile a given case object into a fixed case.

        Args:
            obj: A compiled object.
        Returns:
            The result of compilation.
        Raises:
            CompilationError: when compilation fails.
        """
        return self.compile(obj).fix()

    def compile_default(self, obj: object) -> CaseCompiler:
        """
        Compile a given case object and replace the default value with it.

        Args:
            obj: A compiled object.
        Returns:
            The replaced compiler.
        Raises:
            CompilationError: when compilation fails.
        """
        default = self.compile(obj)
        return self.of_default(default)

    def compile(self, obj: object) -> CaseCompiled:
        """`obj` should be a mapping."""

        obj = ensure_mapping(obj)
        compiled = self._default

        label_obj = obj.get(_KEY_LABEL)
        with on_key(_KEY_LABEL):
            label = ensure_optional_str(label_obj)
            # `label` is always replaced.
            compiled = replace(compiled, label=label)

        enabled_obj = obj.get(_KEY_ENABLED)
        if enabled_obj is not None:
            with on_key(_KEY_ENABLED):
                enabled = ensure_bool(enabled_obj)
            compiled = replace(compiled, enabled=enabled)

        conditions_obj = obj.get(_KEY_CONDITIONS)
        if conditions_obj is not None:
            with on_key(_KEY_CONDITIONS):
                conditions = list(
                    map_compile(
                        self._description.compile,
                        ensure_list(conditions_obj),
                    )
                )
            compiled = replace(compiled, conditions=conditions)

        request_obj = obj.get(_KEY_REQUEST)
        if request_obj is not None:
            with on_key(_KEY_REQUEST):
                request = self._request.compile(request_obj)
            compiled = replace(compiled, request=request)

        response_obj = obj.get(_KEY_RESPONSE)
        if response_obj is not None:
            with on_key(_KEY_RESPONSE):
                response = self._response.compile(response_obj)
            compiled = replace(compiled, response=response)

        wait_obj = obj.get(_KEY_WAIT)
        if wait_obj is not None:
            with on_key(_KEY_WAIT):
                wait = compile_timedelta(wait_obj)
            compiled = replace(compiled, wait=wait)

        return compiled

    def of_default(self, default: CaseCompiled) -> CaseCompiler:
        request = self._request
        if default.request:
            request = self._request.of_default(default.request)

        response = self._response
        if default.response:
            response = self._response.of_default(default.response)

        return CaseCompiler(
            request=request,
            response=response,
            description=self._description,
            default=self._default.replace(default),
        )



from logging import Logger
from typing import Optional

from pluggy import PluginManager
from yamlen import Loader

from preacher.core.logger import default_logger


def create_loader(
    plugin_manager: Optional[PluginManager] = None,
    logger: Optional[Logger] = None,
) -> Loader:
    logger = logger or default_logger

    loader = Loader()
    if plugin_manager:
        plugin_manager.hook.preacher_modify_yaml_loader(loader=loader)
    else:
        logger.debug("No plugin manager is given. The built-in YAML tags are not available.")
    return loader



"""Compilation errors."""

from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from typing import List, Optional, Union, Iterator


@dataclass(frozen=True)
class IndexedNode:
    index: int

    def __str__(self) -> str:
        return f"[{self.index}]"


@dataclass(frozen=True)
class NamedNode:
    name: str

    def __str__(self) -> str:
        return f".{self.name}"


Node = Union[NamedNode, IndexedNode]
Path = List[Node]


def render_path(path: Path) -> str:
    return "".join(str(node) for node in path)


class CompilationError(Exception):
    """Compilation errors."""

    def __init__(
        self,
        message: str,
        node: Optional[Node] = None,
        child: Optional[CompilationError] = None,
        cause: Optional[Exception] = None,
    ):
        self._message = message
        self._node = node
        self._child = child
        self._cause = cause

    @property  # HACK should be cached.
    def path(self) -> Path:
        path = self._child.path if self._child else []
        if self._node:
            path = [self._node] + path
        return path

    def render_path(self) -> str:
        return render_path(self.path)

    def on_node(self, node: Node) -> CompilationError:
        return CompilationError(
            message=self._message,
            node=node,
            child=self,
        )

    def __str__(self) -> str:
        lines = [self._message]
        path = self.path
        if path:
            lines.append(f"  in {render_path(path)}")
        return "\n".join(lines)


@contextmanager
def on_key(key: str) -> Iterator:
    try:
        yield
    except CompilationError as error:
        raise error.on_node(NamedNode(key))


@contextmanager
def on_index(index: int) -> Iterator:
    try:
        yield
    except CompilationError as error:
        raise error.on_node(IndexedNode(index))



from typing import Optional, Type, TypeVar

from preacher.core.context import Context
from preacher.core.value.value import Value

T = TypeVar("T")


class StaticValue(Value[T]):
    def __init__(self, value: T):
        self._value = value

    @property
    def type(self) -> Type[T]:
        return type(self._value)

    def resolve(self, context: Optional[Context] = None) -> T:
        return self._value



"""
Value analysis.
"""

import json
from abc import ABC, abstractmethod
from functools import partial
from typing import Callable, Dict, Generic, Mapping, Optional, TypeVar

from lxml.etree import _Element as Element, XMLParser, fromstring

from preacher.core.request.response import ResponseBody
from preacher.core.util.functional import recursive_map
from preacher.core.util.serialization import to_serializable
from .error import ExtractionError

Source = TypeVar("Source")
T = TypeVar("T")


class Analyzer(ABC):
    """
    Interface to analyze contents.
    """

    @abstractmethod
    def for_text(self, extract: Callable[[str], T]) -> T:
        ...  # pragma: no cover

    @abstractmethod
    def for_mapping(self, extract: Callable[[Mapping], T]) -> T:
        ...  # pragma: no cover

    @abstractmethod
    def for_etree(self, extract: Callable[[Element], T]) -> T:
        ...  # pragma: no cover


class _LazyLoader(Generic[Source, T]):
    """Loads an element tree from a binary content lazily."""

    def __init__(
        self,
        source: Source,
        load: Callable[[Source], T],
        error: Optional[ExtractionError] = None,
    ):
        self._source = source
        self._load = load
        self._error: Exception = error or ExtractionError("Load failed once.")

        self._is_loaded = False
        self._target: Optional[T] = None

    def get(self) -> T:
        """
        Load the target value if needed and returns it.

        Returns:
            The loaded element tree.
        Raises:
            ExtractionError: when loading the target has failed once.
        """
        if not self._is_loaded:
            try:
                self._target = self._load(self._source)
            except Exception:
                pass
            self._is_loaded = True

        target = self._target
        if target is None:
            raise self._error
        return target


JSON_LOAD = json.loads
JSON_ERROR = ExtractionError("Not a valid JSON content")
XML_LOAD = partial(fromstring, parser=XMLParser())
XML_ERROR = ExtractionError("Not a valid XML content")


class ResponseBodyAnalyzer(Analyzer):
    def __init__(self, body: ResponseBody):
        self._body = body
        self._etree_loader = _LazyLoader(body.content, XML_LOAD, XML_ERROR)
        self._json_loader = _LazyLoader(body.text, JSON_LOAD, JSON_ERROR)

    def for_text(self, extract: Callable[[str], T]) -> T:
        return extract(self._body.text)

    def for_mapping(self, extract: Callable[[Mapping], T]) -> T:
        json_value = self._json_loader.get()
        if not isinstance(json_value, Mapping):
            raise ExtractionError(f"Expected a dictionary, but given {type(json_value)}")
        return extract(json_value)

    def for_etree(self, extract: Callable[[Element], T]) -> T:
        return extract(self._etree_loader.get())


def _load_mapping(source: Mapping[str, object]) -> Dict[str, object]:
    return {k: v for k, v in source.items()}


class MappingAnalyzer(Analyzer):
    def __init__(self, value: Mapping[str, object]):
        self._loader = _LazyLoader(value, _load_mapping)

    def for_text(self, extract: Callable[[str], T]) -> T:
        # HACK serialize only one time.
        serializable = recursive_map(to_serializable, self._loader.get())
        return extract(json.dumps(serializable, separators=(",", ":")))

    def for_mapping(self, extract: Callable[[Mapping], T]) -> T:
        return extract(self._loader.get())

    def for_etree(self, extract: Callable[[Element], T]) -> T:
        raise ExtractionError("Not an XML content")



"""Response description compilations."""

from __future__ import annotations

from dataclasses import dataclass, replace
from typing import Optional, List

from preacher.compilation.error import on_key
from preacher.compilation.util.functional import map_compile
from preacher.compilation.util.type import ensure_list, ensure_mapping, or_else
from preacher.core.verification import ResponseDescription, Description, Predicate
from .description import DescriptionCompiler
from .predicate import PredicateCompiler

_KEY_STATUS_CODE = "status_code"
_KEY_HEADERS = "headers"
_KEY_BODY = "body"


@dataclass(frozen=True)
class ResponseDescriptionCompiled:
    status_code: Optional[List[Predicate]] = None
    headers: Optional[List[Description]] = None
    body: Optional[List[Description]] = None

    def replace(
        self,
        other: ResponseDescriptionCompiled,
    ) -> ResponseDescriptionCompiled:
        return ResponseDescriptionCompiled(
            status_code=or_else(other.status_code, self.status_code),
            headers=or_else(other.headers, self.headers),
            body=or_else(other.body, self.body),
        )

    def fix(self) -> ResponseDescription:
        return ResponseDescription(
            status_code=self.status_code,
            headers=self.headers,
            body=self.body,
        )


class ResponseDescriptionCompiler:
    def __init__(
        self,
        predicate: PredicateCompiler,
        description: DescriptionCompiler,
        default: Optional[ResponseDescriptionCompiled] = None,
    ):
        self._predicate = predicate
        self._description = description
        self._default = default or ResponseDescriptionCompiled()

    def compile(self, obj: object) -> ResponseDescriptionCompiled:
        """`obj` should be a mapping."""

        obj = ensure_mapping(obj)
        compiled = self._default

        status_code_obj = obj.get(_KEY_STATUS_CODE)
        if status_code_obj is not None:
            with on_key(_KEY_STATUS_CODE):
                status_code = self._compile_status_code(status_code_obj)
            compiled = replace(compiled, status_code=status_code)

        headers_obj = obj.get(_KEY_HEADERS)
        if headers_obj is not None:
            with on_key(_KEY_HEADERS):
                headers = self._compile_descriptions(headers_obj)
            compiled = replace(compiled, headers=headers)

        body_obj = obj.get(_KEY_BODY)
        if body_obj is not None:
            with on_key(_KEY_BODY):
                body = self._compile_descriptions(body_obj)
            compiled = replace(compiled, body=body)

        return compiled

    def of_default(
        self,
        default: ResponseDescriptionCompiled,
    ) -> ResponseDescriptionCompiler:
        return ResponseDescriptionCompiler(
            predicate=self._predicate,
            description=self._description,
            default=self._default.replace(default),
        )

    def _compile_status_code(self, obj: object) -> List[Predicate]:
        obj = ensure_list(obj)
        return list(map_compile(self._predicate.compile, obj))

    def _compile_descriptions(self, obj: object) -> List[Description]:
        obj = ensure_list(obj)
        return list(map_compile(self._description.compile, obj))



"""CLI Application implementation."""

from typing import Iterable, Optional, Sequence

from preacher.compilation.argument import Arguments
from preacher.compilation.scenario import compile_scenarios
from preacher.compilation.yaml import load_from_paths
from preacher.core.scheduling import create_scheduler
from preacher.core.status import Status
from preacher.plugin.loader import load_plugins
from preacher.plugin.manager import get_plugin_manager
from preacher.presentation.listener import create_listener
from .executor import ExecutorFactory, PROCESS_POOL_FACTORY
from .logging import ColoredFormatter, create_system_logger

__all__ = ["app"]

_REPORT_LOGGER_NAME = "preacher.cli.report.logging"


def app(
    paths: Sequence[str] = (),
    base_url: str = "",
    arguments: Optional[Arguments] = None,
    level: Status = Status.SUCCESS,
    report_dir: Optional[str] = None,
    delay: float = 0.1,
    retry: int = 0,
    timeout: Optional[float] = None,
    concurrency: int = 1,
    executor_factory: Optional[ExecutorFactory] = None,
    plugins: Iterable[str] = (),
    verbosity: int = 0,
) -> int:
    """
    Preacher CLI application.

    Returns:
        the exit code.
    """

    # Fill default.
    arguments = arguments or {}

    logger = create_system_logger(verbosity)

    logger.debug(
        "Running condition\n"
        "  Paths: %s\n"
        "  Arguments: %s\n"
        "  Base URL: %s\n"
        "  Logging report level: %s\n"
        "  Reporting directory path: %s\n"
        "  Max retry count: %d\n"
        "  Delay between attempts in seconds: %s\n"
        "  Timeout in seconds: %s\n"
        "  Concurrency: %s\n"
        "  Executor: %s\n"
        "  Verbosity: %d",
        paths,
        arguments,
        base_url,
        level,
        report_dir,
        retry,
        delay,
        timeout,
        concurrency,
        executor_factory,
        verbosity,
    )

    plugin_manager = get_plugin_manager()
    try:
        load_plugins(plugin_manager, plugins, logger)
    except Exception as error:
        logger.exception(error)
        return 3

    objs = load_from_paths(paths, plugin_manager=plugin_manager, logger=logger)
    scenarios = compile_scenarios(
        objs,
        arguments=arguments,
        plugin_manager=plugin_manager,
        logger=logger,
    )

    listener = create_listener(level=level, formatter=ColoredFormatter(), report_dir=report_dir)
    executor_factory = executor_factory or PROCESS_POOL_FACTORY
    try:
        logger.info("Start running scenarios.")
        with executor_factory.create(concurrency) as executor:
            scheduler = create_scheduler(
                executor=executor,
                listener=listener,
                base_url=base_url,
                timeout=timeout,
                retry=retry,
                delay=delay,
            )
            status = scheduler.run(scenarios)
    except Exception as error:
        logger.exception(error)
        return 3
    finally:
        logger.info("End running scenarios.")

    if not status.is_succeeded:
        return 1

    return 0



"""Request."""

from enum import Enum
from typing import Optional

from .header import Headers
from .request_body import RequestBody
from .url_param import UrlParams


class Method(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"


class Request:
    def __init__(
        self,
        method: Method = Method.GET,
        path: str = "",
        headers: Optional[Headers] = None,
        params: Optional[UrlParams] = None,
        body: Optional[RequestBody] = None,
    ):
        self._method = method
        self._path = path
        self._headers = headers or {}
        self._params = params or {}
        self._body = body

    @property
    def method(self) -> Method:
        return self._method

    @property
    def path(self) -> str:
        return self._path

    @property
    def headers(self) -> Headers:
        return self._headers

    @property
    def params(self) -> UrlParams:
        return self._params

    @property
    def body(self) -> Optional[RequestBody]:
        return self._body



"""Scenario"""

from __future__ import annotations

from typing import List, Optional

from preacher.core.verification import Description
from .case import Case


class Scenario:
    def __init__(
        self,
        label: Optional[str] = None,
        ordered: bool = True,
        conditions: Optional[List[Description]] = None,
        cases: Optional[List[Case]] = None,
        subscenarios: Optional[List[Scenario]] = None,
    ):
        self._label = label
        self._ordered = ordered
        self._conditions = conditions or []
        self._cases = cases or []
        self._subscenarios = subscenarios or []

    @property
    def label(self) -> Optional[str]:
        return self._label

    @property
    def ordered(self) -> bool:
        return self._ordered

    @property
    def conditions(self) -> List[Description]:
        return self._conditions

    @property
    def cases(self) -> List[Case]:
        return self._cases

    @property
    def subscenarios(self) -> List[Scenario]:
        return self._subscenarios



from .request import RequestCompiler
from .request_body import RequestBodyCompiler


def create_request_compiler() -> RequestCompiler:
    body = RequestBodyCompiler()
    return RequestCompiler(body=body)



"""Request."""

from .factory import create_request_compiler
from .request import RequestCompiler, RequestCompiled
from .request_body import RequestBodyCompiler, RequestBodyCompiled
from .url_param import compile_url_params

__all__ = [
    "RequestCompiler",
    "RequestCompiled",
    "RequestBodyCompiler",
    "RequestBodyCompiled",
    "compile_url_params",
    "create_request_compiler",
]



from abc import ABC, abstractmethod
from concurrent.futures import Executor, ProcessPoolExecutor, ThreadPoolExecutor


class ExecutorFactory(ABC):
    """
    Executor factory interface.
    """

    @abstractmethod
    def create(self, concurrency: int) -> Executor:
        """Create an executor."""


class _ProcessPoolFactory(ExecutorFactory):
    def create(self, concurrency: int) -> Executor:
        return ProcessPoolExecutor(concurrency)


class _ThreadPoolFactory(ExecutorFactory):
    def create(self, concurrency: int) -> Executor:
        return ThreadPoolExecutor(concurrency)


PROCESS_POOL_FACTORY = _ProcessPoolFactory()
THREAD_POOL_FACTORY = _ThreadPoolFactory()



from abc import ABC, abstractmethod
from concurrent.futures import Executor
from typing import Iterable, Optional

from requests import Session

from preacher.core.context import Context
from preacher.core.scenario.case import Case
from preacher.core.scenario.case_result import CaseResult
from preacher.core.scenario.case_runner import CaseRunner
from preacher.core.status import StatusedList


class CasesTask(ABC):
    @abstractmethod
    def result(self) -> StatusedList[CaseResult]:
        ...  # pragma: no cover


def _run_cases_in_order(
    runner: CaseRunner,
    cases: Iterable[Case],
    context: Optional[Context],
) -> StatusedList[CaseResult]:
    with Session() as session:
        return StatusedList.collect(
            runner.run(case, session=session, context=context) for case in cases
        )


class OrderedCasesTask(CasesTask):
    def __init__(
        self,
        executor: Executor,
        runner: CaseRunner,
        cases: Iterable[Case],
        context: Optional[Context] = None,
    ):
        self._future = executor.submit(_run_cases_in_order, runner, cases, context)

    def result(self) -> StatusedList[CaseResult]:
        return self._future.result()


class UnorderedCasesTask(CasesTask):
    def __init__(self, executor: Executor, runner: CaseRunner, cases: Iterable[Case]):
        self._futures = [executor.submit(runner.run, case) for case in cases]

    def result(self) -> StatusedList[CaseResult]:
        return StatusedList.collect(f.result() for f in self._futures)



from typing import Iterable, Iterator, Optional

from preacher.core.scenario import Scenario
from preacher.core.scenario import ScenarioRunner
from preacher.core.scenario import ScenarioResult
from preacher.core.scenario import ScenarioTask
from preacher.core.scenario.scenario_task import StaticScenarioTask
from preacher.core.status import Status
from .listener import Listener


class ScenarioScheduler:
    def __init__(self, runner: ScenarioRunner, listener: Optional[Listener] = None):
        self._runner = runner
        self._listener = listener or Listener()

    def run(self, scenarios: Iterable[Scenario]) -> Status:
        """
        Run the scenarios.

        Args:
            scenarios: An iterator of scenarios,
                which can raise `Exception` for each iteration.
        Returns:
            The execution status.
        """
        tasks = self._submit_all(scenarios)
        results = (task.result() for task in list(tasks))

        status = Status.SKIPPED
        for result in results:
            status = status.merge(result.status)
            self._listener.on_scenario(result)

        self._listener.on_end(status)
        return status

    def _submit_all(self, scenarios: Iterable[Scenario]) -> Iterator[ScenarioTask]:
        iterator = iter(scenarios)
        while True:
            try:
                scenario = next(iterator)
            except StopIteration:
                break
            except Exception as error:
                result = ScenarioResult(
                    label="Not a constructed scenario",
                    status=Status.FAILURE,
                    message=f"{error.__class__.__name__}: {error}",
                )
                yield StaticScenarioTask(result)
                continue

            yield self._runner.submit(scenario)



"""
Datetime utilities for Preacher core.
"""

import re
from abc import ABC
from datetime import datetime, time, timedelta, timezone
from time import localtime
from typing import Optional

from dateutil.parser import isoparse

_TIMEDELTA_PATTERN = re.compile(r"([+\-]?\d+)\s*(day|hour|minute|second)s?")


class DatetimeFormat(ABC):
    def format_datetime(self, value: datetime) -> str:
        raise NotImplementedError()

    def parse_datetime(self, value: str) -> datetime:
        raise NotImplementedError()


class Iso8601Format(DatetimeFormat):
    def format_datetime(self, value: datetime) -> str:
        return value.isoformat()

    def parse_datetime(self, value: str) -> datetime:
        try:
            return isoparse(value)
        except ValueError as error:
            raise ValueError(f"An invalid ISO 8601 format: {value}", error)


class StrftimeFormat(DatetimeFormat):
    def __init__(self, format_string: str):
        self._format_string = format_string

    def format_datetime(self, value: datetime) -> str:
        return value.strftime(self._format_string)

    def parse_datetime(self, value: str) -> datetime:
        return datetime.strptime(value, self._format_string)


ISO8601 = Iso8601Format()


class DatetimeWithFormat:
    def __init__(
        self,
        value: datetime,
        fmt: Optional[DatetimeFormat] = None,
    ):
        self._value = value
        self._fmt = fmt or ISO8601

    @property
    def value(self) -> datetime:
        return self._value

    @property
    def fmt(self) -> DatetimeFormat:
        return self._fmt

    @property
    def formatted(self) -> str:
        return self._fmt.format_datetime(self._value)


def now() -> datetime:
    """Returns the current datetime in the system timezone."""
    return datetime.now(system_timezone())


def system_timezone() -> timezone:
    """Returns the system timezone."""
    localtime_ = localtime()
    return timezone(offset=timedelta(seconds=localtime_.tm_gmtoff), name=localtime_.tm_zone)


def parse_time(value: str) -> time:
    """
    Parse a time string.

    Args:
        value: The parsed value.
    Raises:
        ValueError: when given an invalid time string.
    """
    result = time.fromisoformat(value)
    if not result.tzinfo:
        result = result.replace(tzinfo=system_timezone())
    return result


def parse_timedelta(value: str) -> timedelta:
    """
    Parse a timedelta string.

    Args:
        value: The parsed value.
    Raises:
        ValueError: when given an invalid time string.
    """
    normalized = value.strip().lower()
    if not normalized or normalized == "now":
        return timedelta()

    match = _TIMEDELTA_PATTERN.match(normalized)
    if not match:
        raise ValueError(f"Invalid timedelta format: {value}")
    offset = int(match.group(1))
    unit = match.group(2) + "s"
    return timedelta(**{unit: offset})



from logging import Logger, StreamHandler, Formatter, LogRecord, getLogger
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL

from colorama import Fore, Style, init

init()


def _default(message: str) -> str:
    return message


def _info(message: str) -> str:
    return f"{Fore.GREEN}{message}{Style.RESET_ALL}"


def _warning(message: str) -> str:
    return f"{Fore.YELLOW}{message}{Style.RESET_ALL}"


def _error(message: str) -> str:
    return f"{Fore.RED}{message}{Style.RESET_ALL}"


def _critical(message: str) -> str:
    return f"{Fore.RED}{Style.BRIGHT}{message}{Style.RESET_ALL}"


_STYLE_FUNC_MAP = {
    INFO: _info,
    WARNING: _warning,
    ERROR: _error,
    CRITICAL: _critical,
}


class ColoredFormatter(Formatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def format(self, record: LogRecord) -> str:
        style_func = _STYLE_FUNC_MAP.get(record.levelno, _default)
        return style_func(super().format(record))


def create_system_logger(verbosity: int) -> Logger:
    level = _verbosity_to_logging_level(verbosity)
    handler = StreamHandler()
    handler.setLevel(level)
    handler.setFormatter(ColoredFormatter(fmt="[%(levelname)s] %(message)s"))
    logger = getLogger(__name__)
    logger.setLevel(level)
    logger.addHandler(handler)
    return logger


def _verbosity_to_logging_level(verbosity: int) -> int:
    if verbosity > 1:
        return DEBUG
    if verbosity > 0:
        return INFO
    return WARNING



"""Request compilation."""

from .header import Headers
from .request import Request, Method
from .request_body import RequestBody, UrlencodedRequestBody, JsonRequestBody
from .requester import Requester, ExecutionReport, PreparedRequest
from .response import Response, ResponseBody
from .url_param import UrlParams, UrlParam

__all__ = [
    "Request",
    "Method",
    "Headers",
    "UrlParams",
    "UrlParam",
    "RequestBody",
    "UrlencodedRequestBody",
    "JsonRequestBody",
    "Response",
    "ResponseBody",
    "Requester",
    "ExecutionReport",
    "PreparedRequest",
]



import itertools
from logging import Logger
from typing import Optional, Iterator, Iterable

from pluggy import PluginManager

from preacher.compilation.argument import Arguments
from preacher.core.scenario.scenario import Scenario
from .factory import create_scenario_compiler


def compile_scenarios(
    objs: Iterable[object],
    arguments: Optional[Arguments] = None,
    plugin_manager: Optional[PluginManager] = None,
    logger: Optional[Logger] = None,
) -> Iterator[Scenario]:
    compiler = create_scenario_compiler(plugin_manager=plugin_manager, logger=logger)
    return itertools.chain.from_iterable(
        compiler.compile_flattening(obj, arguments=arguments) for obj in objs
    )



"""
Retrying utilities.
"""

import time

from typing import Callable, TypeVar


T = TypeVar("T")


def retry_while_false(
    func: Callable[[], T],
    attempts: int = 1,
    delay: float = 0.1,
    predicate: Callable[[T], bool] = bool,
) -> T:
    if attempts < 1:
        raise ValueError(f"`attempts` must be positive, given {attempts}")

    for _ in range(attempts - 1):
        result = func()
        if predicate(result):
            return result
        time.sleep(delay)

    return func()



import itertools
import sys
from logging import Logger
from typing import Iterator, Optional, Sequence

from pluggy import PluginManager

from preacher.core.logger import default_logger
from .factory import create_loader


def load_from_paths(
    paths: Sequence[str],
    plugin_manager: Optional[PluginManager] = None,
    logger: Optional[Logger] = None,
) -> Iterator[object]:
    logger = logger or default_logger
    loader = create_loader(plugin_manager=plugin_manager, logger=logger)

    if not paths:
        logger.info("No scenario file is given. Load scenarios from stdin.")
        return loader.load_all(sys.stdin)

    return itertools.chain.from_iterable(
        loader.load_all_from_path(_hook_loading(path, logger)) for path in paths
    )


def _hook_loading(path: str, logger: Logger) -> str:
    logger.debug("Load: %s", path)
    return path



from typing import Optional

from yaml import MappingNode, ScalarNode, Node
from yaml.constructor import BaseConstructor
from yamlen import Tag, TagContext
from yamlen.error import on_node

from preacher.compilation.datetime import compile_datetime_format
from preacher.core.datetime import DatetimeFormat, DatetimeWithFormat
from preacher.core.value import Value
from preacher.core.value.impl.datetime import parse_datetime_value_with_format

_KEY_DELTA = "delta"
_KEY_FORMAT = "format"


class RelativeDatetimeTag(Tag):
    def construct(self, node: Node, context: TagContext) -> Value[DatetimeWithFormat]:
        return _construct_relative_datetime(context.constructor, node)


def _construct_relative_datetime(
    constructor: BaseConstructor,
    node: Node,
) -> Value[DatetimeWithFormat]:
    if isinstance(node, ScalarNode):
        return _construct_relative_datetime_of_scalar(constructor, node)
    elif isinstance(node, MappingNode):
        return _construct_relative_datetime_of_mapping(constructor, node)
    else:
        raise ValueError("Invalid relative datetime value format")


def _construct_relative_datetime_of_scalar(
    constructor: BaseConstructor,
    node: ScalarNode,
) -> Value[DatetimeWithFormat]:
    obj = constructor.construct_scalar(node)
    with on_node(node):
        return parse_datetime_value_with_format(obj)


def _construct_relative_datetime_of_mapping(
    constructor: BaseConstructor,
    node: MappingNode,
) -> Value[DatetimeWithFormat]:
    # HACK fix typing.
    format_node: Optional[Node] = None
    datetime_value_node: Optional[Node] = None
    for key_node, value_node in node.value:
        if key_node.value == _KEY_DELTA:
            datetime_value_node = value_node
            continue
        if key_node.value == _KEY_FORMAT:
            format_node = value_node
            continue

    format: Optional[DatetimeFormat] = None
    if format_node:
        obj = constructor.construct_scalar(format_node)  # type: ignore
        with on_node(node):
            format = compile_datetime_format(obj)

    if datetime_value_node:
        datetime_value = constructor.construct_scalar(datetime_value_node)  # type: ignore
        with on_node(datetime_value_node):
            return parse_datetime_value_with_format(datetime_value, format)
    return parse_datetime_value_with_format("now", format)



"""URL Parameters."""

from datetime import date
from typing import Union, List, Mapping, Optional

from preacher.core.context import Context
from preacher.core.datetime import DatetimeWithFormat
from preacher.core.value import Value

UrlParam = Union[object, List[object]]
UrlParams = Union[str, Mapping[str, UrlParam]]
ResolvedUrlParamValue = Optional[str]
ResolvedUrlParam = Union[ResolvedUrlParamValue, List[ResolvedUrlParamValue]]
ResolvedUrlParams = Union[str, Mapping[str, ResolvedUrlParam]]


def resolve_url_param_value(value: object, context: Optional[Context] = None) -> Optional[str]:
    """
    Resolve a URL parameter value.

    Args:
        value: A URL parameter value, which is one of the items below.
            - A `bool` value, which is resolved into "true" or "false".
            - A `int` value.
            - A `float` value.
            - A `str` value.
            - A `date` value, which is resolved into its ISO format.
            - A `DatetimeWithFormat` value, which is resolved into its formatted string.
            - A value of one of them.
        context: A resolution context.
    Returns:
        The resolved value.
    Raises:
        ValueError: when given a not resolvable value.
    """
    if isinstance(value, Value):
        value = value.resolve(context)

    if value is None:
        return None
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, DatetimeWithFormat):
        return value.formatted
    if isinstance(value, date):
        return value.isoformat()
    if isinstance(value, (str, int, float)):
        return str(value)
    raise ValueError(f"Invalid URL parameter value: {value}")


def resolve_url_param(param: UrlParam, context: Optional[Context] = None) -> ResolvedUrlParam:
    """
    Resolve a URL parameter.

    Args:
        param: A URL parameter, which is one of the items below.
            - A URL parameter value.
            - A list of URL parameter values.
            - A Value of a list of URL parameter values.
        context: A resolution context.
    Returns:
        The resolved value.
    Raises:
        ValueError: when given a not resolvable value.
    """
    if isinstance(param, Value):
        param = param.resolve(context)

    if isinstance(param, list):
        return [resolve_url_param_value(value, context) for value in param]
    return resolve_url_param_value(param, context)


def resolve_url_params(params: UrlParams, context: Optional[Context] = None) -> ResolvedUrlParams:
    if isinstance(params, str):
        return params
    return {key: resolve_url_param(param, context) for (key, param) in params.items()}



from logging import Logger
from typing import Optional

from pluggy import PluginManager

from preacher.compilation.extraction import ExtractionCompiler, create_extraction_compiler
from preacher.core.logger import default_logger
from .description import DescriptionCompiler
from .matcher import MatcherFactoryCompiler
from .predicate import PredicateCompiler
from .response import ResponseDescriptionCompiler


def create_matcher_factory_compiler(
    plugin_manager: Optional[PluginManager] = None,
    logger: Optional[Logger] = None,
) -> MatcherFactoryCompiler:
    logger = logger or default_logger

    compiler = MatcherFactoryCompiler()
    if plugin_manager:
        plugin_manager.hook.preacher_add_matchers(compiler=compiler)
    else:
        logger.info("No plugin manager is given. The default matchers are not available.")
    return compiler


def create_predicate_compiler(
    matcher_factory: Optional[PluginManager] = None,
    plugin_manager: Optional[PluginManager] = None,
    logger: Optional[Logger] = None,
) -> PredicateCompiler:
    matcher_factory = matcher_factory or create_matcher_factory_compiler(
        plugin_manager=plugin_manager,
        logger=logger,
    )
    return PredicateCompiler(matcher_factory)


def create_description_compiler(
    extraction: Optional[ExtractionCompiler] = None,
    predicate: Optional[PredicateCompiler] = None,
    plugin_manager: Optional[PluginManager] = None,
    logger: Optional[Logger] = None,
) -> DescriptionCompiler:
    extraction = extraction or create_extraction_compiler(
        plugin_manager=plugin_manager,
        logger=logger,
    )
    predicate = predicate or create_predicate_compiler(
        plugin_manager=plugin_manager,
        logger=logger,
    )
    return DescriptionCompiler(extraction=extraction, predicate=predicate)


def create_response_description_compiler(
    predicate: Optional[PredicateCompiler] = None,
    description: Optional[DescriptionCompiler] = None,
    plugin_manager: Optional[PluginManager] = None,
    logger: Optional[Logger] = None,
) -> ResponseDescriptionCompiler:
    predicate = predicate or create_predicate_compiler(
        plugin_manager=plugin_manager, logger=logger
    )
    description = description or create_description_compiler(
        predicate=predicate,
        plugin_manager=plugin_manager,
        logger=logger,
    )
    return ResponseDescriptionCompiler(predicate=predicate, description=description)



from yamlen import Loader
from yamlen.tag.impl.inclusion import InclusionTag

from .argument import ArgumentTag
from .context import ContextTag
from .datetime import RelativeDatetimeTag

__all__ = ["add_default_tags"]


def add_default_tags(loader: Loader) -> None:
    loader.add_tag("!include", InclusionTag())
    loader.add_tag("!argument", ArgumentTag())
    loader.add_tag("!context", ContextTag())
    loader.add_tag("!relative_datetime", RelativeDatetimeTag())



from .case import CaseCompiler, CaseCompiled
from .factory import create_scenario_compiler
from .integration import compile_scenarios
from .scenario import ScenarioCompiler

__all__ = [
    "CaseCompiler",
    "CaseCompiled",
    "ScenarioCompiler",
    "create_scenario_compiler",
    "compile_scenarios",
]



from logging import Formatter
from typing import Optional

from preacher.core.scheduling import Listener, MergingListener
from preacher.core.status import Status
from .logging import create_logging_reporting_listener
from .html import create_html_reporting_listener


def create_listener(
    level: Status = Status.SUCCESS,
    formatter: Optional[Formatter] = None,
    report_dir: Optional[str] = None,
) -> Listener:
    merging = MergingListener()
    merging.append(create_logging_reporting_listener(level=level, formatter=formatter))
    if report_dir:
        merging.append(create_html_reporting_listener(report_dir))
    return merging



from abc import ABC, abstractmethod
from functools import partial
from typing import Any, Callable, Iterator, Optional

from preacher.core.extraction import Extractor, Analyzer
from preacher.core.util.functional import identity, apply_if_not_none


class JqEngine(ABC):
    @abstractmethod
    def iter(self, query: str, value: str) -> Iterator[object]:
        ...  # pragma: no cover


class JqExtractor(Extractor):
    def __init__(
        self,
        engine: JqEngine,
        query: str,
        multiple: bool = False,
        cast: Optional[Callable[[object], Any]] = None,
    ):
        self._engine = engine
        self._query = query
        self._multiple = multiple
        self._cast = cast or identity

    def extract(self, analyzer: Analyzer) -> object:
        values = (
            apply_if_not_none(self._cast, value)
            for value in analyzer.for_text(partial(self._engine.iter, self._query))
        )
        if self._multiple:
            return list(values)
        else:
            return next(values, None)



"""CLI Options."""

import logging
import re
import shlex
from enum import IntEnum
from typing import Iterable, Mapping, Optional, Tuple, Union, Any

from click import Choice
from click import Context
from click import BadParameter
from click import Parameter
from click import ParamType
from click import Option
from click import STRING

from yaml import safe_load
from yaml.error import MarkedYAMLError

from preacher.compilation.argument import Arguments
from preacher.core.status import Status
from .executor import ExecutorFactory, PROCESS_POOL_FACTORY, THREAD_POOL_FACTORY


class Level(IntEnum):
    SKIPPED = logging.DEBUG
    SUCCESS = logging.INFO
    UNSTABLE = logging.WARN
    FAILURE = logging.ERROR

    def __str__(self):
        return self.name.lower()


_LEVEL_MAP: Mapping[str, Level] = {str(level): level for level in Level}
_CONCURRENT_EXECUTOR_FACTORY_MAP: Mapping[str, ExecutorFactory] = {
    "process": PROCESS_POOL_FACTORY,
    "thread": THREAD_POOL_FACTORY,
}


class ArgumentType(ParamType):

    _base = STRING
    name = _base.name

    def convert(self, value, param, ctx):
        exp = self._base.convert(value, param, ctx)
        return _parse_argument(exp)

    def split_envvar_value(self, rv):
        return shlex.split(rv)


class LevelType(ParamType):

    _choice = Choice(tuple(item.name.lower() for item in Status), case_sensitive=False)
    name = _choice.name

    def get_metavar(self, param: Parameter) -> str:
        return self._choice.get_metavar(param)

    def get_missing_message(self, param: Parameter) -> str:
        return self._choice.get_missing_message(param)

    def convert(self, value: Any, param: Optional[Parameter], ctx: Optional[Context]) -> Status:
        if isinstance(value, Status):
            return value
        key = self._choice.convert(value, param, ctx).lower()
        return next(item for item in Status if item.name.lower() == key)


class ExecutorFactoryType(ParamType):

    _choice = Choice(
        tuple(_CONCURRENT_EXECUTOR_FACTORY_MAP.keys()),
        case_sensitive=False,
    )
    name = _choice.name

    def get_metavar(self, param):
        return self._choice.get_metavar(param)

    def get_missing_message(self, param):
        return self._choice.get_missing_message(param)

    def convert(
        self,
        value: Any,
        param: Optional[Parameter],
        ctx: Optional[Context],
    ) -> ExecutorFactory:
        if isinstance(value, ExecutorFactory):
            return value
        key = self._choice.convert(value, param, ctx)
        return _CONCURRENT_EXECUTOR_FACTORY_MAP[key.lower()]


def pairs_callback(
    _context: Context,
    _option_or_parameter: Union[Option, Parameter],
    value: Iterable[Tuple[str, object]],
) -> Arguments:
    return dict(value)


def positive_float_callback(
    _context: Context,
    _option_or_parameter: Union[Option, Parameter],
    value: Optional[float],
) -> Optional[float]:
    if value is None:
        return value

    if value <= 0.0:
        raise BadParameter(f"must be positive, given {value}")
    return value


def _parse_argument(value: str) -> Tuple[str, object]:
    match = re.match(r"^([^=]+)=(.*)$", value)
    if not match:
        raise BadParameter(f"Invalid format argument: {value}")

    key = match.group(1)
    try:
        value = safe_load(match.group(2))
    except MarkedYAMLError as error:
        raise BadParameter(f"Invalid YAML format: {value}\n{error}")

    return key, value



import os
from typing import Iterable

import jinja2

from preacher.core.request import Response, ExecutionReport
from preacher.core.scenario import ScenarioResult


class HtmlReporter:
    def __init__(self, path):
        self._path = path
        self._responses_path = os.path.join(self._path, "responses")

        self._loader = jinja2.PackageLoader("preacher", "resources/report/html")

        self._initialize()

    def _initialize(self) -> None:
        os.makedirs(self._path, exist_ok=True)
        os.makedirs(self._responses_path, exist_ok=True)

    def export_response(self, execution: ExecutionReport, response: Response) -> None:
        name = f"{response.id}.html"
        path = os.path.join(self._responses_path, name)

        env = jinja2.Environment(loader=self._loader, autoescape=True)
        template = env.get_template("response-view.html")
        with open(path, "w") as f:
            template.stream(execution=execution, response=response).dump(f)

    def export_results(self, results: Iterable[ScenarioResult]) -> None:
        html_path = os.path.join(self._path, "index.html")

        env = jinja2.Environment(loader=self._loader, autoescape=True)
        template = env.get_template("index.html")
        with open(html_path, "w") as f:
            template.stream(scenarios=results).dump(f)



"""Context definitions."""

from contextlib import contextmanager
from typing import Iterator, MutableMapping

CONTEXT_KEY_STARTS = "starts"
CONTEXT_KEY_BASE_URL = "base_url"


class Context(MutableMapping[str, object]):
    def __init__(self, **kwargs):
        self._inner: MutableMapping[str, object] = dict(**kwargs)

    def __setitem__(self, key: str, value: object) -> None:
        self._inner[key] = value

    def __delitem__(self, key: str) -> None:
        del self._inner[key]

    def __getitem__(self, key: str) -> object:
        return self._inner[key]

    def __len__(self) -> int:
        return len(self._inner)

    def __iter__(self) -> Iterator[str]:
        return iter(self._inner)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Context):
            return False
        return self._inner == other._inner


@contextmanager
def closed_context(context: Context, **kwargs) -> Iterator[Context]:
    stored = {key: context[key] for key in kwargs if key in context}

    context.update(kwargs)
    yield context

    for key in kwargs:
        if key in stored:
            context[key] = stored[key]
        else:
            if key in context:
                del context[key]



from yaml import Node
from yamlen import Tag, TagContext

from preacher.core.value.impl.context import ContextualValue


class ContextTag(Tag):
    def construct(self, node: Node, context: TagContext) -> ContextualValue:
        key = context.constructor.construct_scalar(node)  # type: ignore
        return ContextualValue(key)



from .factory import create_listener
from .html import HtmlReportingListener, create_html_reporting_listener
from .logging import LoggingReportingListener, create_logging_reporting_listener

__all__ = [
    "HtmlReportingListener",
    "LoggingReportingListener",
    "create_html_reporting_listener",
    "create_logging_reporting_listener",
    "create_listener",
]



"""
Test cases, which execute a given request and verify its response
along the given descriptions.
"""

from datetime import timedelta
from typing import Optional, List

from preacher.core.request import Request
from preacher.core.verification import Description
from preacher.core.verification import ResponseDescription


class Case:
    """
    Test cases, which execute a given request and verify its response
    along the given descriptions.
    """

    def __init__(
        self,
        label: Optional[str] = None,
        enabled: bool = True,
        conditions: Optional[List[Description]] = None,
        request: Optional[Request] = None,
        response: Optional[ResponseDescription] = None,
        waiting_time: Optional[timedelta] = None,
    ):
        self._label = label
        self._enabled = enabled
        self._conditions = conditions or []
        self._request = request or Request()
        self._response = response or ResponseDescription()
        self._waiting_time = waiting_time or timedelta()

    @property
    def label(self) -> Optional[str]:
        return self._label

    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def conditions(self) -> List[Description]:
        return self._conditions

    @property
    def request(self) -> Request:
        return self._request

    @property
    def response(self) -> ResponseDescription:
        return self._response

    @property
    def waiting_time(self) -> timedelta:
        return self._waiting_time



from preacher.core.scheduling.factory import create_scheduler
from preacher.core.scheduling.listener import Listener, MergingListener
from preacher.core.scheduling.scenario_scheduler import ScenarioScheduler

__all__ = [
    "ScenarioScheduler",
    "Listener",
    "MergingListener",
    "create_scheduler",
]



from time import sleep
from typing import Optional

import requests

from preacher.core.context import Context, closed_context
from preacher.core.datetime import now
from preacher.core.extraction import MappingAnalyzer
from preacher.core.unit import UnitRunner
from preacher.core.verification import Verification
from .case import Case
from .case_listener import CaseListener
from .case_result import CaseResult


class CaseRunner:
    def __init__(self, unit_runner: UnitRunner, listener: Optional[CaseListener] = None):
        self._unit_runner = unit_runner
        self._listener = listener or CaseListener()

    @property
    def base_url(self) -> str:
        return self._unit_runner.base_url

    def run(
        self,
        case: Case,
        session: Optional[requests.Session] = None,
        context: Optional[Context] = None,
    ) -> CaseResult:
        if not case.enabled:
            return CaseResult(label=case.label)

        context = context if context is not None else Context()
        with closed_context(context, starts=now(), base_url=self.base_url) as context:
            context_analyzer = MappingAnalyzer(context)
            conditions = Verification.collect(
                condition.verify(context_analyzer, context) for condition in case.conditions
            )
            if not conditions.status.is_succeeded:
                return CaseResult(case.label, conditions)

            sleep(max(case.waiting_time.total_seconds(), 0.0))

            execution, response, verification = self._unit_runner.run(
                request=case.request,
                requirements=case.response,
                session=session,
                context=context,
            )
        self._listener.on_execution(execution, response)

        return CaseResult(case.label, conditions, execution, verification)



from typing import List, Optional

from preacher.core.request import Response, ExecutionReport
from preacher.core.scenario import ScenarioResult
from preacher.core.scheduling import Listener
from preacher.core.status import Status
from preacher.presentation.html import HtmlReporter


class HtmlReportingListener(Listener):
    def __init__(self, reporter: HtmlReporter):
        self._reporter = reporter
        self._results: List[ScenarioResult] = []

    def on_execution(self, execution: ExecutionReport, response: Optional[Response]) -> None:
        if not response:
            return
        self._reporter.export_response(execution, response)

    def on_scenario(self, result: ScenarioResult) -> None:
        self._results.append(result)

    def on_end(self, status: Status) -> None:
        self._reporter.export_results(self._results)


def create_html_reporting_listener(path: str):
    reporter = HtmlReporter(path)
    return HtmlReportingListener(reporter)



from datetime import date

from preacher.core.datetime import DatetimeWithFormat


def to_serializable(value: object) -> object:
    if isinstance(value, DatetimeWithFormat):
        return value.formatted
    if isinstance(value, date):
        return value.isoformat()
    return value



"""Request compilation."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import Optional, Tuple

from preacher.compilation.argument import Argument, Arguments, inject_arguments
from preacher.compilation.error import CompilationError, on_key
from preacher.compilation.util.type import ensure_str, ensure_mapping, or_else
from preacher.core.request import Request, Method, UrlParams
from .request_body import RequestBodyCompiled, RequestBodyCompiler
from .url_param import compile_url_params

_KEY_METHOD = "method"
_KEY_PATH = "path"
_KEY_HEADERS = "headers"
_KEY_PARAMS = "params"
_KEY_BODY = "body"

_METHOD_MAP = {method.name: method for method in Method}


@dataclass(frozen=True)
class RequestCompiled:
    method: Optional[Method] = None
    path: Optional[str] = None
    headers: Optional[Mapping[str, str]] = None
    params: Optional[UrlParams] = None
    body: Optional[RequestBodyCompiled] = None

    def replace(self, other: RequestCompiled) -> RequestCompiled:
        return RequestCompiled(
            method=or_else(other.method, self.method),
            path=or_else(other.path, self.path),
            headers=or_else(other.headers, self.headers),
            params=other.params if other.params is not None else self.params,
            body=or_else(other.body, self.body),
        )

    def fix(self) -> Request:
        return Request(
            method=or_else(self.method, Method.GET),
            path=or_else(self.path, ""),
            headers=self.headers,
            params=self.params,
            body=self.body.fix() if self.body else None,
        )


class RequestCompiler:
    def __init__(
        self,
        body: RequestBodyCompiler,
        default: RequestCompiled = None,
    ):
        self._body = body
        self._default = default or RequestCompiled()

    def compile(
        self,
        obj: object,
        arguments: Optional[Arguments] = None,
    ) -> RequestCompiled:
        """
        Compiles an object into an intermediate request.

        Args:
            obj: A compiled object, which should be a mapping or a string.
            arguments: Arguments to inject.
        Returns:
            The result of compilation.
        Raises:
            CompilationError: when compilation fails.
        """
        if isinstance(obj, Argument):
            obj = inject_arguments(obj, arguments)

        if isinstance(obj, str):
            return self.compile({_KEY_PATH: obj}, arguments)

        obj = ensure_mapping(obj)
        compiled = self._default

        method_obj = obj.get(_KEY_METHOD)
        if method_obj is not None:
            with on_key(_KEY_METHOD):
                method_obj = inject_arguments(method_obj, arguments)
                method = _compile_method(method_obj)
            compiled = replace(compiled, method=method)

        path_obj = obj.get(_KEY_PATH)
        if path_obj is not None:
            with on_key(_KEY_PATH):
                path_obj = inject_arguments(path_obj, arguments)
                path = ensure_str(path_obj)
            compiled = replace(compiled, path=path)

        headers_obj = obj.get(_KEY_HEADERS)
        if headers_obj is not None:
            with on_key(_KEY_HEADERS):
                headers_obj = inject_arguments(headers_obj, arguments)
                headers = _compile_headers(headers_obj)
            compiled = replace(compiled, headers=headers)

        params_obj = obj.get(_KEY_PARAMS)
        if params_obj is not None:
            with on_key(_KEY_PARAMS):
                params = compile_url_params(params_obj, arguments)
            compiled = replace(compiled, params=params)

        body_obj = obj.get(_KEY_BODY)
        if body_obj is not None:
            with on_key(_KEY_BODY):
                body = self._body.compile(body_obj, arguments)
            compiled = replace(compiled, body=body)

        return compiled

    def of_default(self, default: RequestCompiled) -> RequestCompiler:
        body = self._body
        if default.body:
            body = body.of_default(default.body)

        return RequestCompiler(body=body, default=self._default.replace(default))


def _compile_method(obj: object) -> Method:
    key = ensure_str(obj).upper()
    method = _METHOD_MAP.get(key)
    if not method:
        message = f"Must be in {list(_METHOD_MAP)}, but given: {obj}"
        raise CompilationError(message)
    return method


def _compile_headers(obj: object) -> Mapping[str, str]:
    obj = ensure_mapping(obj)
    items = (_compile_header_item(k, v) for (k, v) in obj.items())
    return dict(filter(None, items))


def _compile_header_item(key: object, value: object) -> Optional[Tuple[str, str]]:
    if value is None:
        return None
    return ensure_str(key), ensure_str(value)



import logging
from typing import Iterable, Optional

from preacher.core.scenario import ScenarioResult
from preacher.core.scheduling import Listener
from preacher.core.status import Status
from preacher.presentation.logging import LoggingReporter, create_logging_reporter


class LoggingReportingListener(Listener):
    def __init__(self, reporter: LoggingReporter):
        self._reporter = reporter

    def on_scenario(self, result: ScenarioResult) -> None:
        self._reporter.show_scenario_result(result)

    def on_end(self, status: Status) -> None:
        self._reporter.show_status(status)


def create_logging_reporting_listener(
    reporter: Optional[LoggingReporter] = None,
    logger: Optional[logging.Logger] = None,
    logger_name: str = "",
    level: Status = Status.SUCCESS,
    handlers: Optional[Iterable[logging.Handler]] = None,
    formatter: Optional[logging.Formatter] = None,
) -> LoggingReportingListener:
    """
    Create a logging reporting listener.

    Args:
        reporter: A reporter. When given this, the other parameters are ignored.
        logger: A logger where reports logged.
            When given this, `logger_name`, `level`, `handlers` and `formatter` are ignored.
        logger_name: The logger name. When not given, it will be automatically generated.
        level: The minimum level to report.
        handlers: The logging handlers. When given, `formatter` is ignored.
        formatter: The logging formatter.
    """
    reporter = reporter or create_logging_reporter(
        logger=logger,
        logger_name=logger_name,
        level=level,
        handlers=handlers,
        formatter=formatter,
    )
    return LoggingReportingListener(reporter)
